/*
FEU SERVIR EL COMPILADOR D'ESP32 v2.0.17 . La part de CAN no funciona amb versions 3
  ESP32-WROOM-32D with wifi antenna.

  I2C_ADDR 0x2A

* Register:
b7 b6 b5 b4 b3 b2 b1 b0
x  x  x  x  0  0  0  1     REG_RD
x  x  x  x  0  0  1  0     REG_WR
0  0  0  0  0  1  1  0     LCB_VERS

** Register REG_RD:
b7 b6 b5 b4 b3 b2 b1 b0
x  x  x  x  0  0  0  1     REG_RD


** Register VERSION:
b7 b6 b5 b4 b3 b2 b1 b0
x  x  x  x  x  1  1  0     VERT_VERS

Raspberry Pi needs two I2C sendings to read properly.
i2cget -y 1 0x2A 0x06 w
0x2030
i2cget -y 1 0x2A 0x06 w
0x0100

Answer 0x0100:
01: VERSION
00: SUBVERSION  
*/
#if CONFIG_FREERTOS_UNICORE
#define ARDUINO_RUNNING_CORE 0
#else
#define ARDUINO_RUNNING_CORE 1
#endif

#include "driver/twai.h"

#include <WiFi.h>
#include "IoT-02_wifiMng.h"
#include "IoT-02_mqttCredentials.h"
#include "IoT-02_mqttTopics.h"
#include <PubSubClient.h>
#include <ESPmDNS.h>
#include <Wire.h>

#include <ArduinoJson.h>
#include "EncoderStepCounter.h"

#define ENCODER_PIN1 33
#define ENCODER_INT1 digitalPinToInterrupt(ENCODER_PIN1)
#define ENCODER_PIN2 34
#define ENCODER_INT2 digitalPinToInterrupt(ENCODER_PIN2)

// #define CAN_PREFIX_LED_BT 0x100

#define N_SJD 500

#define MINIMUM_DELAY_ALLOWING_MULTITASKING 50

#define I2C_ADDR 0x2A /* PCF8574 addr: 0x20 .. 0x27*/

#define LCB_VERSION    0x01
#define LCB_SUBVERSION 0x00

WiFiClient espClient;
PubSubClient client(espClient);
#define MAC_SIZE 15
char sMac[MAC_SIZE];

TaskHandle_t TaskMqtt;

StaticJsonDocument<N_SJD> doc, docPub; // https://arduinojson.org/

// Create instance for one full step encoder
EncoderStepCounter encoder(ENCODER_PIN1, ENCODER_PIN2);
// Use the following for half step encoders
//EncoderStepCounter encoder(ENCODER_PIN1, ENCODER_PIN2, HALF_STEP);

signed long position = 0;

#define CTRL_ESP 2
#define MQTT_LED 32
#define WEB 23

#define BLUE_LED      MQTT_LED
#define YELLOW_LED    MQTT_LED
#define BT_IO0 0

#define I2C_SDA 21
#define I2C_SCL 22

#define RX_PIN        26
#define TX_PIN        27
//#define CAN_RS        38

#define CANPID_RPM    0x0C
#define CAN_REQST_ID  0x7DF
#define CAN_REPLY_ID  0x7E8

HardwareSerial rPiSer(2);

#define REG_RD    0x01
#define REG_WR    0x02
#define LCB_VERS  0x06

TwoWire I2C_M = TwoWire(0); // Main I2C. ESP32-S3 acting as slave
byte byLastReadNumber;

void onRequest(){ 
  word wAnswer;

  Serial.println("onRequest");
  Serial.print("byLastReadNumber: 0x");
  Serial.println(byLastReadNumber,HEX);
  
  if((byLastReadNumber & 0x0F) == REG_RD){
    I2C_M.write((const uint8_t*)&position, sizeof(long));
    /*
    if(!(byLastReadNumber & 0x80)){ // b7: 0 - Byte level
      I2C_M.write(byLeds);
      Serial.print("Esclau respon 0x");
      Serial.print(byLeds,HEX);
    }else{ // b7: 1 - Bit level
      I2C_M.write((byLeds & (byte)(1 << (0x07 & (byLastReadNumber>>4))))?0x01:0x00);
      Serial.print("Esclau respon 0x");
      Serial.print((byLeds & (byte)(1 << (0x07 & (byLastReadNumber>>4))))?0x01:0x00,HEX);
    } */
  }
  if((byLastReadNumber & 0x0F) == LCB_VERS){
    I2C_M.write(LCB_SUBVERSION);
    I2C_M.write(LCB_VERSION);
    Serial.print("Esclau respon 0x");
    Serial.print(LCB_SUBVERSION,HEX);
    Serial.print(" 0x");
    Serial.println(LCB_VERSION,HEX);
  }
}

void onReceive(int len){
  enum {rd,wr,versio,no} enReg = no;
  byte byReg,byWhichDev = 0x00;
  int n = 0;
  short int siN;
  
  Serial.printf("onReceive[%d]: ", len);

  // digitalWrite(LED_BRD,HIGH);
  
  while (1 < I2C_M.available()) { // loop through all but the last
    byLastReadNumber = I2C_M.read();   // receive byte as a character
    Serial.print(n,DEC);
    Serial.print(") El màster m'ha fet arribar un 0x");
    Serial.println(byLastReadNumber,HEX);
    /*
    switch(n){
      case 0:
        Serial.print("byLastReadNumber & 0x07: 0x");
        Serial.println(byLastReadNumber & 0x07,HEX);
        switch(byLastReadNumber & 0x07){
          case REG_RD:  enReg = rd; Serial.println("enReg = rd");break;
          case REG_WR:  enReg = wr; Serial.println("enReg = wr");    break;
          case LCB_VERS: enReg = versio; Serial.println("enReg = versio");break;
          default: enReg = no; Serial.println("enReg = no");
        }
        byReg = byLastReadNumber;
        break;
    }
    */
    n++;
  }
  byLastReadNumber = I2C_M.read();
  Serial.print(n,DEC);
  Serial.print("> El màster m'ha fet arribar un 0x");
  Serial.println(byLastReadNumber,HEX);
  /*
  switch(n){
    case 1:
      switch(enReg){
        case wr:
          if(!(byLastReadNumber & 0x80)){ // b7: 0 - Byte level
            byLeds = byLastReadNumber;
            Serial.print("byLeds: 0x");
            Serial.println(byLeds,HEX);            
          }else{ // b7: 1 - Bit level
            if( byLastReadNumber & 0x01 ){
              byLeds |= (byte)(1 << (0x07 & (byLastReadNumber>>4)));
            }else{
              byLeds &= ~(byte)(1 << (0x07 & (byLastReadNumber>>4)));
            }
            Serial.print("Sending ");
            Serial.print( byLastReadNumber & 0x01 );
            Serial.print(" to button led ");
            Serial.print((0x07 & (byLastReadNumber>>4)) + 1);
            Serial.print(" - byLeds: 0x");
            Serial.println(byLeds,HEX);
          }
          break;
      }
      break;
  }
  */
  // digitalWrite(LED_BRD,LOW);
}

unsigned int uiConverter(unsigned int uiN) {
  unsigned int uiOut = 0x0000;

  // Serial.print("uiN: 0x");
  // Serial.println(uiN,HEX);
  uiOut |= (uiN & 0x0001) ? 0x0010 : 0x0000;
  uiOut |= (uiN & 0x0002) ? 0x0008 : 0x0000;
  uiOut |= (uiN & 0x0004) ? 0x0004 : 0x0000;
  uiOut |= (uiN & 0x0008) ? 0x0002 : 0x0000;
  uiOut |= (uiN & 0x0010) ? 0x0001 : 0x0000;
  uiOut |= (uiN & 0x0020) ? 0x1000 : 0x0000;
  uiOut |= (uiN & 0x0040) ? 0x0800 : 0x0000;
  uiOut |= (uiN & 0x0080) ? 0x0400 : 0x0000;
  uiOut |= (uiN & 0x0100) ? 0x0200 : 0x0000;
  uiOut |= (uiN & 0x0200) ? 0x0100 : 0x0000;
  uiOut |= (uiN & 0x0400) ? 0x2000 : 0x0000;
  // Serial.print("uiOut: 0x");
  // Serial.println(uiOut,HEX);
  return uiOut;
}
/*
void vSetupI2cRelayModule(byte byAddr) {
  Wire.beginTransmission(byAddr);
  Wire.write(0x04);
  Wire.write(0x00);
  Wire.write(0x00);
  Wire.endTransmission();
  Wire.beginTransmission(byAddr);
  Wire.write(0x06);
  Wire.write(0x00); // OUTPUT
  Wire.write(0x00); // OUTPUT
  Wire.endTransmission();
}

void vWriteRelayModule(byte byAddr, unsigned int uiV) {
  byte byChipEnable = 0xA0;

  Wire.beginTransmission(byAddr);
  Wire.write(0x02);
  Wire.write(byte(byChipEnable | (0xFF & uiConverter(uiV))));
  Wire.write(byte(uiConverter(uiV) >> 8));
  Wire.endTransmission();
  Serial.print("uiA1: 0x");
  Serial.println(byte(byChipEnable | (0xFF & uiConverter(uiV))), HEX);
  Serial.print("uiA2: 0x");
  Serial.println(byte(uiConverter(uiV) >> 8), HEX);
  Wire.beginTransmission(byAddr);
  Wire.write(0x02);
  Wire.write(byte(0xFF & uiConverter(uiV)));
  Wire.write(byte(uiConverter(uiV) >> 8));
  Wire.endTransmission();
  Serial.print("uiA3: 0x");
  Serial.println(byte(0xFF & uiConverter(uiV)), HEX);
  Serial.print("uiA4: 0x");
  Serial.println(byte(uiConverter(uiV) >> 8), HEX);
  Serial.println("---------");
}

void vWriteInputsModule(byte byAddr, unsigned int uiV) {
  Wire.beginTransmission(byAddr);
  Wire.write(0x02);
  Wire.write(byte(0xFF & uiV));
  Wire.write(byte(uiV >> 8));
  Wire.endTransmission();
}
*/
// Call tick on every change interrupt
void interrupt() {
  encoder.tick();
}

void mqttconnect() {
  const TickType_t xDelay = 5000 / portTICK_PERIOD_MS; // 5 seconds

  /* Loop until reconnected */
  while (!client.connected()) {
    Serial.print("Head MAC Address:  ");
    Serial.println(sMac);
    Serial.print("MQTT connecting ...");
    /* client ID */
    String clientId = "IoT-02_" + String(sMac); // <-------   Unique name in every device
    /* connect now */
    if (client.connect(clientId.c_str(), mqtt_user, mqtt_password)) {
      //if (client.connect(clientId.c_str())) {
      Serial.println("connected");
      /* subscribe topic */
      client.subscribe(TOPIC_REQUEST_MAC); // <-------   Subscription to MQTT(S) topic
      /*client.subscribe(String("/" + String(sMac) + TOPIC_LED_W).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_LED_R).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_LED_Y).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_LED_G).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_LATCHING_RELAY).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_REQUEST_LATCHING_RELAY_STATE).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_REQUEST_LDR).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_REQUEST_T).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_REQUEST_RH).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_REQUEST_P).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_REQUEST_ALT).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_SMALL_TEXT).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_MEDIUM_TEXT).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_BIG_TEXT).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_JSON_INPUT_REQ).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_JSON_LEDS).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_JSON_BME_REQ).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_SYNC_LDR).c_str());
        client.subscribe(String("/" + String(sMac) + TOPIC_SYNC_JSONIN).c_str());
        Serial.println("Subscriptions to: ");
        Serial.println(String("/" + String(sMac) + TOPIC_SYNC_LDR).c_str());
        Serial.println(String("/" + String(sMac) + TOPIC_SYNC_JSONIN).c_str());*/
      client.subscribe(String("/" + String(sMac) + TOPIC_CAN_JSON_MSG_IN).c_str());
    } else {
      Serial.print("failed, status code =");
      Serial.print(client.state());
      Serial.println("try again in 5 seconds");
      /* Wait 5 seconds before retrying */
      //delay(5000);
      vTaskDelay(xDelay);
    }
  }
}

/*
      Serial.println((word)doc["ID"],HEX);
      Serial.printf("{ ID: 0x%X\n",(word)doc["ID"]);
      int n = (byte)doc["n"];
      for(int i = 0; i < n; i++){
        Serial.printf("[%d]: 0x%X, ",i,(byte)doc["msg"][i]);
      }
      Serial.println("}");
*/


int nCANsend () {
  static unsigned int n = 250;
  static byte byOutA = 0xFF, byOutB = 0x00, byMask = 0x01;

  n++;

  byOutA = ~byMask;
  byOutB = byMask;
  byMask = (byMask == 0x80) ? 0x01 : byMask << 1;

  twai_message_t message;
  message.identifier = (word)doc["ID"];
  message.extd = (byte)doc["extd"];
  message.rtr = (byte)doc["rtr"];
  message.data_length_code = (byte)doc["n"];
  for (int i = 0; i < message.data_length_code; i++) {
    message.data[i] = (byte)doc["msg"][i];
  }
  /* message.data[0] = 2;
    message.data[1] = 1;
    message.data[2] = CANPID_RPM;
    message.data[3] = byOutA;
    message.data[4] = byOutB;
    message.data[5] = 0;
    message.data[6] = (byte)(n & 0xFF);
    message.data[7] = (byte)(n >> 8);
    //return twai_transmit(&message, 0);*/

  if (twai_transmit(&message, pdMS_TO_TICKS(1000)) == ESP_OK) {
    digitalWrite(MQTT_LED,HIGH);
    Serial.println("Message queued for transmission");

    if (message.extd) {
      Serial.println("Message is in Extended Format\n");
    } else {
      Serial.println("Message is in Standard Format\n");
    }
    Serial.printf("nCANsend() --> ID is 0x%X\n", message.identifier);
    if (!(message.rtr)) {
      for (int i = 0; i < message.data_length_code; i++) {
        Serial.printf("Data byte %d = 0x%X\n", i, message.data[i]);
      }
    }
    digitalWrite(MQTT_LED,LOW);

    return 1;
  } else {
    Serial.println("Failed to queue message for transmission");
    return 0;
  }
}

void receivedCallback(char* topic, byte* payload, unsigned int length) {
  String szTopic = String(topic), szPayload = "";
  char topicBuf[250];

  Serial.print("Topic: ");
  Serial.println(topic);

  Serial.print("payload: ");
  for (int i = 0; i < length; i++) {
    Serial.print((char)payload[i]);
    szPayload += (char)payload[i];
  }
  Serial.println();

  if (szTopic == String("/" + String(sMac) + TOPIC_CAN_JSON_MSG_IN).c_str()) {
    DeserializationError error = deserializeJson(doc, payload);
    if (error) {
      Serial.print(F("deserializeJson() failed: "));
      Serial.println(error.f_str());
      return;
    } else {
      Serial.println((word)doc["ID"], HEX);
      Serial.printf("{ ID: 0x%X\n", (word)doc["ID"]);
      int n = (byte)doc["n"];
      if ((byte)doc["rtr"] == 0) { // Dades
        for (int i = 0; i < n; i++) {
          Serial.printf("[%d]: 0x%X, ", i, (byte)doc["msg"][i]);
        }
      } else {
        Serial.println("Asking for an answer");
      }
      Serial.println("}");
      nCANsend();
    }
  }
}

void vSetupMqtt() {
  /* set SSL/TLS certificate */
  // espClient.setCACert(ca_cert);
  /* configure the MQTT server with IPaddress and port */
  client.setServer(mqtt_server, mqtt_port);
  /* this receivedCallback function will be invoked
    when client received subscribed topic */
  client.setCallback(receivedCallback);
}

void vReconnectWifiMqtt() {
  vSetupWifi();
  szGetMac().toCharArray(sMac, MAC_SIZE);
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
  vSetupMqtt();
  mqttconnect();
}

void vTaskMqtt( void * pvParameters ) {
  byte byI2cAddr;

  Serial.print("vTaskMqtt running on core ");
  Serial.println(xPortGetCoreID());
  for (;;) {
    if (!client.connected()) {
      vReconnectWifiMqtt();
    }
    client.loop();
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);
  }
  vTaskDelete(NULL); // There is an infinite loop before. This line will never be reached.
}

boolean bPressedButton(int nWhichOne) {
  if (digitalRead(nWhichOne))
    return false;
  return true;
}

int CANsend (void) {
  static unsigned int n = 250;
  static byte byOutA = 0xFF, byOutB = 0x00, byMask = 0x01;

  n++;

  byOutA = ~byMask;
  byOutB = byMask;
  byMask = (byMask == 0x80) ? 0x01 : byMask << 1;

  twai_message_t message;
  message.identifier = CAN_REQST_ID;
  message.extd = 0;
  message.rtr = 0;
  message.data_length_code = 8;
  message.data[0] = 2;
  message.data[1] = 1;
  message.data[2] = CANPID_RPM;
  message.data[3] = byOutA;
  message.data[4] = byOutB;
  message.data[5] = 0;
  message.data[6] = (byte)(n & 0xFF);
  message.data[7] = (byte)(n >> 8);
  //return twai_transmit(&message, 0);

  if (twai_transmit(&message, pdMS_TO_TICKS(1000)) == ESP_OK) {
    Serial.println("Message queued for transmission");

    if (message.extd) {
      Serial.println("Message is in Extended Format\n");
    } else {
      Serial.println("Message is in Standard Format\n");
    }
    Serial.printf("CANsend(void) --> ID is 0x%X\n", message.identifier);
    if (!(message.rtr)) {
      for (int i = 0; i < message.data_length_code; i++) {
        Serial.printf("Data byte %d = 0x%X\n", i, message.data[i]);
      }
    }

    return 1;
  } else {
    Serial.println("Failed to queue message for transmission");
    return 0;
  }
}

void CANrecv (void) {
  char output[N_SJD];

  twai_message_t message;

  if (twai_receive(&message, 0) != ESP_OK) {
    return;
  }
  digitalWrite(MQTT_LED, HIGH);
  Serial.println("CANrecv{");
  // digitalWrite(BLUE_LED,HIGH);
  //delay(20);
  //digitalWrite(BLUE_LED, LOW);

  //Process received message
  if (message.extd) {
    Serial.println("Message is in Extended Format\n");
  } else {
    Serial.println("Message is in Standard Format\n");
  }
  Serial.printf("message.identifier & 0x00FF = 0x%X\n", message.identifier & 0x00FF);
  docPub["ID"] = (word)message.identifier;
  docPub["extd"] = (byte)message.extd;
  docPub["rtr"] = (byte)message.rtr;
  docPub["n"] = (byte)message.data_length_code;

  if (!(message.rtr)) {
    for (int i = 0; i < message.data_length_code; i++) {
      Serial.printf("Data byte %d = 0x%X\n", i, (byte)message.data[i]);
      docPub["msg"][i] = (byte)message.data[i];
    }
  }
  Serial.println("}CANrecv");
  serializeJson(docPub, output);
  client.publish( String("/" + String(sMac) + TOPIC_CAN_JSON_MSG_OUT).c_str(), output);
  Serial.println("Mqtt publishes:");
  Serial.print("Topic: ");
  Serial.println(String("/" + String(sMac) + TOPIC_CAN_JSON_MSG_OUT).c_str());
  Serial.print("Payload: ");
  Serial.println(output);
  rPiSer.println(output);
  digitalWrite(MQTT_LED,LOW);
}

void vTaskRpi(void *pvParameters) {
  (void) pvParameters;
  // https://stackoverflow.com/questions/55698070/sending-json-over-serial-in-python-to-arduino#55744475
  for (;;) {
    int     size_ = 0;
    String  payload;
    while ( !rPiSer.available()  ) {}
    if ( rPiSer.available() )
      payload = rPiSer.readStringUntil( '\n' );
    // StaticJsonDocument<512> doc;
    // doc, declarat com a global, es fa servir per a nCANsend();

    /* DeserializationError   error = deserializeJson(doc, payload);
    if (error) {
      Serial.println(error.c_str());
      // return;
    } else {
      if (doc["operation"] == "sequence") {
        // rPiSer.println("{\"Success\":\"True\"}");
        rPiSer.printf("{\"Encoder\":%d}\n\r",position);
      }
      else {
        rPiSer.println("{\"Position\":\"False\"}");
      }*/
      DeserializationError error = deserializeJson(doc, payload);
      if (error) {
        Serial.print(F("deserializeJson() failed: "));
        Serial.println(error.f_str());
        //return;
      } else {
        bool bCanMsgBroadcast = doc.containsKey("ID") & doc.containsKey("n") & doc.containsKey("rtr") & doc.containsKey("msg");
        bool bCanMsgLedButtonCommand = doc.containsKey("ID") & doc.containsKey("button") & doc.containsKey("value") & doc.containsKey("rtr");
        if(bCanMsgBroadcast || bCanMsgLedButtonCommand){
          // if(bCanMsgLedButtonCommand){ Ja no val perquè escric doc["msg"]
          if((word)doc["ID"] & 0x100){ // CAN_PREFIX_LED_BT 0x100
            /*
            data["ID"] = addr | 0x100 # CAN_PREFIX_LED_BT
            data["button"] = button
            data["value"] = value
            */
            Serial.println((word)doc["ID"], HEX);
            if ((byte)doc["rtr"] == 0) { // Dades
              int nBt = int(doc["button"]);
              int nVal = int(doc["value"]);
              Serial.printf("{ ID: 0x%X, button: %d, value: %d\n", (word)doc["ID"],nBt,nVal);
              doc["n"] = 2;
              doc["msg"][0] = nBt;
              doc["msg"][1] = nVal;
            } else {
              Serial.println("Asking for an answer");
            }
            Serial.println("}");
          }else{
            Serial.println((word)doc["ID"], HEX);
            Serial.printf("{ ID: 0x%X\n", (word)doc["ID"]);
            int n = (byte)doc["n"];
            if ((byte)doc["rtr"] == 0) { // Dades
              for (int i = 0; i < n; i++) {
                Serial.printf("[%d]: 0x%X, ", i, (byte)doc["msg"][i]);
              }
            } else {
              Serial.println("Asking for an answer");
            }
            Serial.println("}");
          }

          nCANsend();
        }
        bool bReq = doc.containsKey("req");
        if(bReq){
          if(doc["req"] == "encoder"){
            rPiSer.printf("{\"Encoder\":%d}\n\r",position);
            Serial.println("Es demana encoder");
          }
        }
      } 
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);  // one tick delay (15ms) in between reads for stability
  }
}

void vTaskEncoder(void *pvParameters) {
  (void) pvParameters;

  // Initialize encoder
  encoder.begin();
  // Initialize interrupts
  attachInterrupt(ENCODER_INT1, interrupt, CHANGE);
  attachInterrupt(ENCODER_INT2, interrupt, CHANGE);
  for (;;) {
    static bool sb = HIGH, bSurtoWeb = false;
    static unsigned int uiR = 0x0400;
    signed char pos = encoder.getPosition();

    /*
    if (!digitalRead(WEB)) {
      bSurtoWeb = true;
      vWriteRelayModule(I2C_ADDR_R, uiR);
      delay(200);
      if (uiR == 0x0400) {
        uiR = 0x0001;
      } else {
        uiR <<= 1;
      }
    } else {
      if (bSurtoWeb) {
        bSurtoWeb = false;
        vWriteRelayModule(I2C_ADDR_R, 0x0000);
        uiR = 0x0400;
      }
      // digitalWrite(MQTT_LED, sb = !sb);
      // delay((sb)?100:900);
    }
    */
    if (pos != 0) {
      position += pos;
      Serial.print("pos: ");
      Serial.print(pos);
      encoder.reset();
      Serial.print(", position: ");
      Serial.println(position);
      // digitalWrite(MQTT_LED, sb = !sb);
    }
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);  // one tick delay (15ms) in between reads for stability
  }
}

void vTaskReadingCan(void *pvParameters) {
  (void) pvParameters;

  // vSetupCan();

  for (;;) {
    CANrecv();
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);  // one tick delay (15ms) in between reads for stability
  }
}

void setup() {
  // Wire.begin(I2C_SDA, I2C_SCL);
  Serial.begin(115200);
  Serial.println(__FILE__);
  rPiSer.begin(115200);

  Serial.print("I2C address: 0x");
  Serial.println(I2C_ADDR,HEX);
  I2C_M.onReceive(onReceive);
  I2C_M.onRequest(onRequest);
  I2C_M.begin(I2C_ADDR,I2C_SDA,I2C_SCL,400000);

  // Serial.setTxTimeoutMs(0); // prevent slow serial prints if no usb
  pinMode(BT_IO0, INPUT);
  pinMode(CTRL_ESP, OUTPUT);
  pinMode(MQTT_LED, OUTPUT);
  pinMode(WEB, INPUT_PULLUP);
  digitalWrite(CTRL_ESP, LOW);
  delay(1000); // Temps per a configurar tots els mòduls
  digitalWrite(CTRL_ESP, HIGH);
  // vWriteRelayModule(I2C_ADDR_R, 0x0000);
  // pinMode(CAN_RS, OUTPUT);    // INPUT (high impedance) = slope control mode, OUTPUT = see next line
  // digitalWrite(CAN_RS, LOW);  // LOW = high speed mode, HIGH = low power mode (listen only)
  twai_general_config_t g_config = TWAI_GENERAL_CONFIG_DEFAULT((gpio_num_t)TX_PIN, (gpio_num_t)RX_PIN, TWAI_MODE_NORMAL);  // TWAI_MODE_NORMAL, TWAI_MODE_NO_ACK or TWAI_MODE_LISTEN_ONLY
  // twai_timing_config_t t_config  = TWAI_TIMING_CONFIG_500KBITS();
  twai_timing_config_t t_config  = TWAI_TIMING_CONFIG_100KBITS();
  twai_filter_config_t f_config  = TWAI_FILTER_CONFIG_ACCEPT_ALL();
  twai_driver_install(&g_config, &t_config, &f_config);
  twai_start();

  // xTaskCreatePinnedToCore(
  //   vTaskMqtt,   /* Task function. */
  //   "vTaskMqtt",     /* name of task. */
  //   10000,       /* Stack size of task */
  //   NULL,        /* parameter of the task */
  //   1,           /* priority of the task */
  //   &TaskMqtt,      /* Task handle to keep track of created task */
  //   0);          /* pin task to core 0 */

  xTaskCreatePinnedToCore(
    vTaskEncoder
    ,  "vTaskEncoder"   // A name just for humans
    ,  2 * 1024 // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  3  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  ARDUINO_RUNNING_CORE);

  xTaskCreatePinnedToCore(
    vTaskRpi
    ,  "vTaskRpi"   // A name just for humans
    ,  2 * 1024 // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  2  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  ARDUINO_RUNNING_CORE);
    
  xTaskCreatePinnedToCore(
    vTaskReadingCan
    ,  "vTaskReadingCan"   // A name just for humans
    ,  8 * 1024 // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  3  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  ARDUINO_RUNNING_CORE);    
}

void loop() {
  static boolean bIO0wasPressed = false;
  boolean bIO0currentState = bPressedButton(BT_IO0);
  static unsigned int n = 1;

  if (bIO0wasPressed != bIO0currentState) {
    delay(2);
    if (bIO0currentState) {
      Serial.print(n++);
      Serial.println(") Button 0 pressed (BOOT)");
      CANsend();
    }
    bIO0wasPressed = bIO0currentState;
  }
  // CANrecv(); --> Reading CAN in a task

  vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);
}
